/*
 *	T 3r0c -- Copyright (C) 1997 Nils M. Holm
 *	<fs29@rumms.uni-mannheim.de>
 *
 *	T is a compiler for a simple procedural language (T).
 *	It reads a source program from stdin and writes code
 *	described in `c.h' to stdout.
 *	The compiler uses an ultra-simple code generator which
 *	produces code that simulates a stack machine.
 *
 *	The T compiler is free software. Do what you want with it,
 *	but do not remove any copyright notices.
 *
 *	The risk of using T is entirely with you, so don't complain.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "c.h"


#define MAXSYMB		1024
#define NSPACE		3072
#define TEXTLEN		128
#define MAXDEPTH	32
#define MAXTBL		128
#define META		0x100

#define CNST		1
#define GLOBAL		2
#define FUNCT		4
#define PROTO		8

#define iswhite(C)	((C) == ' ' || (C) == '\t' || (C) == '\n' ||\
			(C) == '\r' || (C) == '\f')

struct sym {
	char	*name;
	int	flags;
	int	size;
	int	val;
} Symbols[MAXSYMB];
int	St = 0;
int	Sym_equ, Sym_mod;
char	Names[NSPACE];
int	Nt = 0;
int	Line = 1;
int	Errcount = 0;
int	Token;
char	Text[TEXTLEN+1];
int	Value;
int	Op;
int	Label = 1;
int	Startlab = 0;
int	Endlab = 0;
int	Exitlab = 0;
int	Looplevl = 0;
int	Locladdr = 1;


#define ENDFILE	-1
#define ERROR	4095

#define SYMBOL	0
#define NUMBER	1
#define STRING	2

#define	BINOP	10
#define	DISOP	11
#define CONOP	12
#define UNOP	20
#define ADDROP	21
#define LPAREN	30
#define RPAREN	31
#define LBRACK	40
#define RBRACK	41
#define SEMI	50
#define COMMA	60
#define ASSIGN	70
#define COND	80
#define COLON	81

struct oper {
	int	prec;
	int	len;
	char	*name;
	int	token;
	int	(*op)();
} Ops[] = {
	{	6, 1, "+",	BINOP,	add	},
	{	7, 2, "*",	BINOP,	multiply},
	{	0, 1, ";",	SEMI,	NULL	},
	{	0, 1, ",",	COMMA,	NULL	},
	{	0, 1, "(",	LPAREN,	NULL	},
	{	0, 1, ")",	RPAREN,	NULL	},
	{	0, 1, "[",	LBRACK,	NULL	},
	{	0, 1, "]",	RBRACK,	NULL	},
	{	3, 1, "=",	BINOP,	equ	},
	{	5, 1, "&",	BINOP,	binand	},
	{	5, 1, "|",	BINOP,	binor	},
	{	5, 1, "^",	BINOP,	binxor	},
	{	0, 1, "@",	ADDROP,	NULL	},
	{	0, 1, "~",	UNOP,	binnot	},
	{	0, 1, ":",	COLON,	NULL	},
	{	0, 2, ":=",	ASSIGN,	NULL	},
	{	0, 1, "\\",	UNOP,	lognot	},
	{	1, 2, "\\/",	DISOP,	NULL	},
	{	3, 2, "\\=",	BINOP,	nequ	},
	{	4, 1, "<",	BINOP,	less	},
	{	4, 2, "<=",	BINOP,	lteq	},
	{	5, 2, "<<",	BINOP,	binshl	},
	{	4, 1, ">",	BINOP,	grtr	},
	{	4, 2, ">=",	BINOP,	gteq	},
	{	5, 2, ">>",	BINOP,	binshr	},
	{	6, 1, "-",	BINOP,	subtract},
	{	0, 2, "->",	COND,	NULL	},
	{	7, 2, "/",	BINOP,	divide	},
	{	2, 2, "/\\",	CONOP,	NULL	},
	{	7, 1, "mod",	BINOP,	modulo	},
	{	0, 0, NULL,	0,	NULL	}
};

#define KCONST	100
#define KDECL	101
#define KDO	102
#define KELSE	103
#define KEND	104
#define KFOR	105
#define KHALT	106
#define KIE	107
#define KIF	108
#define KLEAVE	109
#define KLOOP	110
#define KRETURN	111
#define KVAR	112
#define KWHILE	113


error(m, s)
char	*m, *s;
{
	if (s)
		fprintf(stderr, "%d: %s: %s\n", Line, m, s);
	else
		fprintf(stderr, "%d: %s\n", Line, m);
	Errcount++;
}


fatal(m, s)
char	*m, *s;
{
	error(m, s);
	fprintf(stderr, "terminating.\n");
	exit(1);
}


getce() {
	int	c;

	c = getchar();
	if (c != '\\') return(c);
	c = getchar();
	if (c == 'a') return('\a');
	else if (c == 'b') return('\b');
	else if (c == 'e') return('\033');
	else if (c == 'f') return('\f');
	else if (c == 'n') return('\n');
	else if (c == 'q') return('"'|META);
	else if (c == 'r') return('\r');
	else if (c == 's') return(' ');
	else if (c == 't') return('\t');
	else if (c == 'v') return('\v');
	else return(c);
}


findkw(s)
char	*s;
{
	switch(*s) {
		case 'c':
			if (!strcmp(s, "const")) return(KCONST);
			return(0);
		case 'd':
			if (!strcmp(s, "decl")) return(KDECL);
			if (!strcmp(s, "do")) return(KDO);
			return(0);
		case 'e':
			if (!strcmp(s, "else")) return(KELSE);
			if (!strcmp(s, "end")) return(KEND);
			return(0);
		case 'f':
			if (!strcmp(s, "for")) return(KFOR);
			return(0);
		case 'h':
			if (!strcmp(s, "halt")) return(KHALT);
			return(0);
		case 'i':
			if (!strcmp(s, "ie")) return(KIE);
			if (!strcmp(s, "if")) return(KIF);
			return(0);
		case 'l':
			if (!strcmp(s, "leave")) return(KLEAVE);
			if (!strcmp(s, "loop")) return(KLOOP);
			return(0);
		case 'm':
			if (!strcmp(s, "mod")) return(BINOP);
			return(0);
		case 'r':
			if (!strcmp(s, "return")) return(KRETURN);
			return(0);
		case 'v':
			if (!strcmp(s, "var")) return(KVAR);
			return(0);
		case 'w':
			if (!strcmp(s, "while")) return(KWHILE);
			return(0);
		default:
			return(0);
	}
}


findop(c) {
	int	i = 0, j;

	while (Ops[i].len) {
		if (c == *Ops[i].name) {
			if (Ops[i+1].len > 1) {
				j = i++;
				c = getchar();
				while (Ops[i].len == 2) {
					if (c == Ops[i].name[1]) {
						Op = i;
						return(Ops[i].token);
					}
					i++;
				}
				ungetc(c, stdin);
				Op = j;
				return(Ops[j].token);
			}
			else {
				Op = i;
				return(Ops[i].token);
			}
		}
		i++;
	}
	return(0);
}


findop2(s)
char	*s;
{
	int	i = 0;

	while (Ops[i].len) {
		if (!strcmp(Ops[i].name, s))
			return(i);
		i++;
	}
	fatal("bad text in findop2()", s);
}


scan() {
	int	c, i;
	char	bc[3];

	c = getchar();
	for (;;) {
		while (iswhite(c)) {
			if (c == '\n') Line++;
			c = getchar();
		}
		if (c != '!') break;
		while (c != '\n' && !feof(stdin))
			c = getchar();
	}
	if (feof(stdin)) return(ENDFILE);
	if (isalpha(c) || c == '_') {
		i = 1;
		Text[0] = '_';
		while (isalpha(c) || isdigit(c) || c == '_') {
			if (isupper(c)) c = tolower(c);
			if (i == TEXTLEN-1) error("symbol too long", NULL);
			if (i < TEXTLEN-1) Text[i] = c;
			i++;
			c = getchar();
		}
		Text[i] = 0;
		ungetc(c, stdin);
		c = findkw(Text+1);
		if (c) {
			if (c == BINOP) Op = Sym_mod;
			return(c);
		}
		Text[i++] = '_';
		Text[i] = 0;
		return(SYMBOL);
	}
	else if (isdigit(c) || c == '%') {
		Value = 0;
		i = 0;
		if (c == '%') {
			i = 1;
			c = getchar();
		}
		while (isdigit(c)) {
			Value = Value*10 + (c-'0');
			c = getchar();
		}
		ungetc(c, stdin);
		if (i) Value = -Value;
		return(NUMBER);
	}
	else if (c == '\'') {
		Value = getce() & ~META;
		if (getchar() != '\'') error("missing `''", NULL);
		return(NUMBER);
	}
	else if (c == '"') {
		i = 0;
		c = getce();
		while (c != '"') {
			if (i == TEXTLEN) error("string too long", NULL);
			if (i < TEXTLEN) Text[i] = c & ~META;
			i++;
			c = getce();
			if (feof(stdin)) fatal("unexpected EOF", NULL);
		}
		Text[i] = 0;
		return(STRING);
	}
	else {
		i = findop(c);
		if (i) return(i);
		sprintf(bc, "%02x", c);
		fatal("bad input character", bc);
	}
}


findsym(name)
char	*name;
{
	int	i;

	for (i=St-1; i>=0; i--) {
		if (!strcmp(name, Symbols[i].name))
			return(i);
	}
	return(-1);
}


newsym(name, size, val, flags)
char	*name;
int	size, val;
{
	int	k = strlen(name)+1, sp;

	if (*name != '*' && (sp = findsym(name)) != -1) {
		if (flags != FUNCT || Symbols[sp].flags != (FUNCT|PROTO)) {
			error("duplicate symbol", name);
			return(sp);
		}
	}
	if (St >= MAXSYMB) fatal("too many symbols", name);
	if (Nt + k >= NSPACE) fatal("out of symbol space", name);
	strcpy(Names+Nt, name);
	Symbols[St].name = Names+Nt;
	Nt += k;
	Symbols[St].size = size;
	Symbols[St].val = val;
	Symbols[St].flags = flags;
	return(St++);
}


newlab() {
	if (!Label) fatal("out of labels", NULL);
	return(Label++);
}


match(t, m)
int	t;
char	*m;
{
	if (t == Token)
		Token = scan();
	else
		error(m, NULL);
}


semi() {
	match(SEMI, "missing `;'");
}


lparen() {
	match(LPAREN, "missing `('");
}


rparen() {
	match(RPAREN, "missing `)'");
}


equsign() {
	if (Token != BINOP || Op != Sym_equ)
		error("missing `='", NULL);
	else
		Token = scan();
}


synch(t) {
	if (t) {
		while (Token != t) {
			if ((Token = scan()) == EOF)
				fatal("EOF found in error recovery", NULL);
		}
	}
	else {
		Token = scan();
	}
}


constval() {
	int	k, v = 0;

	if (Token == NUMBER) {
		v = Value;
		Token = scan();
	}
	else if (Token == SYMBOL) {
		if ((k = findsym(Text)) == -1) {
			error("undefined symbol", Text);
			v = 1;
		}
		else {
			if (!(Symbols[k].flags & CNST))
				error("not a constant", Text);
			else
				v = Symbols[k].val;
		}
		Token = scan();
	}
	else {
		error("constant value expected", NULL);
	}
	return(v);
}


fncall(k) {
	int	n = 0;

	if (k == -1) error("call of non-function", NULL);
	if (k >= 0 && !(Symbols[k].flags & FUNCT))
		error("call of non-function", Symbols[k].name);
	lparen();
	if (Token != RPAREN) for (;;) {
		expr();
		n++;
		if (Token != COMMA) break;
		Token = scan();
	}
	if (k >= 0 && Symbols[k].size != n)
		error("wrong number of arguments", Symbols[k].name);
	if (k >= 0) call(Symbols[k].name, n);
	rparen();
}


loadsym(k) {
	if (Symbols[k].flags & FUNCT) return;
	if (Symbols[k].flags & GLOBAL)
		loadglob(Symbols[k].name, Symbols[k].val, Symbols[k].size);
	else
		loadlocl(Symbols[k].val, Symbols[k].size);
}


savesym(k) {
	if (Symbols[k].flags & GLOBAL)
		saveglob(Symbols[k].name, Symbols[k].val);
	else
		savelocl(Symbols[k].val);
}


address(lval) {
	int	k;

	if (Token != SYMBOL) {
		error("symbol expected", NULL);
		return(-1);
	}
	if ((k = findsym(Text)) == -1)
		error("undefined symbol", Text);
	if (k != -1 && Symbols[k].flags & CNST) {
		numlit(Symbols[k].val);
	}
	else {
		if (!lval && k != -1) loadsym(k);
	}
	Token = scan();
	while (Token == LBRACK) {
		if (k >= 0 && (Symbols[k].flags & (CNST|FUNCT)))
			error("bad subscript", Symbols[k].name);
		if (lval) {
			if (k == -1)
				indirect();
			else
				loadsym(k);
		}
		Token = scan();
		expr();
		deref(lval);
		match(RBRACK, "missing `]'");
		k = -1;
	}
	return(k);
}


table() {
	int	t[MAXTBL], tf[MAXTBL], p = 0;
	int	v, i;

	Token = scan();
	while (	Token == SYMBOL || Token == NUMBER ||
		Token == LBRACK || Token == STRING
	) {
		if (Token == SYMBOL || Token == NUMBER) {
			t[p] = constval();
			tf[p] = 0;
		}
		else if (Token == STRING) {
			t[p] = strlit(Text);
			tf[p] = 1;
			Token = scan();
		}
		else {
			t[p] = table();
			tf[p] = 2;
		}
		if (p++ >= MAXTBL) fatal("table too big", NULL);
		if (Token != COMMA) break;
		Token = scan();
	}
	v = sveclit();
	for (i=0; i<p; i++) vecmem(t[i], tf[i]);
	eveclit();
	match(RBRACK, "missing `]'");
	return(v);
}


factor() {
	struct oper	*o;
	int		k;

	if (Token == NUMBER) {
		numlit(Value);
		Token = scan();
	}
	else if (Token == SYMBOL) {
		k = address(0);
		if (Token == LPAREN) {
			fncall(k);
		}
		else {
			if (k >= 0 && (Symbols[k].flags & FUNCT))
				error("incomplete function call",
					Symbols[k].name);
		}
	}
	else if (Token == STRING) {
		loadlab(strlit(Text));
		Token = scan();
	}
	else if (Token == LBRACK) {
		loadlab(table());
	}
	else if (Token == ADDROP) {
		Token = scan();
		k = address(1);
		if (k >= 0) {
			if (Symbols[k].flags & CNST)
				error("cannot take @const", Symbols[k].name);
			if (Symbols[k].flags & GLOBAL)
				loadglob(Symbols[k].name, Symbols[k].val, 1);
			else
				loadlocl(Symbols[k].val, 1);
		}
	}
	else if (Token == BINOP) {
		if (*Ops[Op].name != '-' || Ops[Op].name[1])
			error("bad unary operator", NULL);
		Token = scan();
		factor();
		negate();
	}
	else if (Token == UNOP) {
		o = &Ops[Op];
		Token = scan();
		factor();
		(*o->op)();
	}
	else if (Token == LPAREN) {
		Token = scan();
		expr();
		rparen();
	}
	else {
		error("bad expression", NULL);
		synch(0);
	}
}


emitop(stk, sp)
int	*stk;
int	sp;
{
	struct oper	*o;

	o = &Ops[stk[sp-1]];
	(*o->op)();
	sp--;
	return(sp);
}


binary() {
	int	stk[MAXDEPTH], sp = 0;
	int	fac = 0;
	struct oper	*o;

	factor();
	for (;;) {
		if (Token == BINOP) {
			if (fac) {
				error("missing factor", NULL);
				sp--;
			}
			while (sp && Ops[Op].prec <= (&Ops[stk[sp-1]])->prec)
					sp = emitop(stk, sp);
			if (sp >= MAXDEPTH) fatal("expr too complex", NULL);
			stk[sp++] = Op;
			fac = 1;
			Token = scan();
		}
		else {
			if (fac) {
				factor();
				fac = 0;
			}
			else {
				while (sp) sp = emitop(stk, sp);
				return;
			}
		}
	}
}


conj() {
	int	e = 0;

	binary();
	if (Token == CONOP) e = newlab();
	while (Token == CONOP) {
		Token = scan();
		tcond(e, 2);
		binary();
	}
	if (e) econd(e);
}


disj() {
	int	e = 0;

	conj();
	if (Token == DISOP) e = newlab();
	while (Token == DISOP) {
		Token = scan();
		tcond(e, 1);
		conj();
	}
	if (e) econd(e);
}


expr() {
	int	y = newlab(), n = newlab();

	disj();
	if (Token == COND) {
		Token = scan();
		tcond(n, 0);
		expr();
		fcond(y, n);
		match(COLON, "missing `:'");
		expr();
		econd(y);
	}
}


ie_stmt(alt) {
	int	y, n;

	n = newlab();
	if (alt) y = newlab();
	Token = scan();
	lparen();
	expr();
	rparen();
	tcond(n, 0);
	stmt();
	if (alt) {
		fcond(y, n);
		match(KELSE, "missing `ELSE'");
		stmt();
	}
	econd(alt? y: n);
}


while_stmt() {
	int	s = Startlab, e = Endlab;
	int	levl = Looplevl;

	Startlab = newlab();
	Endlab = newlab();
	Looplevl = Locladdr;
	Token = scan();
	lparen();
	mark(Startlab);
	expr();
	tcond(Endlab, 0);
	rparen();
	stmt();
	jump(Startlab);
	econd(Endlab);
	Startlab = s;
	Endlab = e;
	Looplevl = levl;
}


for_stmt() {
	int	k, step = 1;
	int	r, s = Startlab, e = Endlab;
	int	levl = Looplevl;

	Startlab = newlab();
	r = newlab();
	Endlab = newlab();
	Looplevl = Locladdr;
	Token = scan();
	lparen();
	if (Token == SYMBOL) {
		if ((k = findsym(Text)) < 0)
			error("undefined symbol", Text);
		if (k != -1 && Symbols[k].flags & (CNST|FUNCT))
			error("bad lvalue", Text);
		if (k != -1 && Symbols[k].size)
			error("index may not be a vector", Text);
	}
	match(SYMBOL, "symbol name expected");
	equsign();
	expr();
	if (k != -1) savesym(k);
	match(COMMA, "missing `,'");
	mark(r);
	if (k != -1) loadsym(k);
	expr();
	if (Token == COMMA) {
		Token = scan();
		step = constval();
	}
	rparen();
	forcond(step, Endlab);
	stmt();
	mark(Startlab);
	if (Symbols[k].flags & GLOBAL)
		incrglob(Symbols[k].name, Symbols[k].val, step);
	else
		incrlocl(Symbols[k].val, step);
	jump(r);
	mark(Endlab);
	Startlab = s;
	Endlab = e;
	Looplevl = levl;
}


asg_or_call() {
	int	k;

	if (Token != SYMBOL) {
		error("bad statement", NULL);
		synch(SEMI);
		return;
	}
	k = address(1);
	if (Token == ASSIGN) {
		Token = scan();
		expr();
		if (k != -1) {
			if (Symbols[k].size)
				error("missing subscript", Symbols[k].name);
			if (Symbols[k].flags & (CNST|FUNCT))
				error("bad lvalue", Symbols[k].name);
			savesym(k);
		}
		else {
			storevec();
		}
	}
	else if (Token == LPAREN) {
		fncall(k);
		pop();
	}
	else {
		error("bad statement", NULL);
		synch(SEMI);
	}
}


stmt() {
	if (Token == KDO) {
		compound();
	}
	else if (Token == KIE) {
		ie_stmt(1);
	}
	else if (Token == KIF) {
		ie_stmt(0);
	}
	else if (Token == KWHILE) {
		while_stmt();
	}
	else if (Token == KFOR) {
		for_stmt();
	}
	else if (Token == KLEAVE) {
		Token = scan();
		semi();
		if (!Endlab) error("`LEAVE' in wrong context", NULL);
		unstack(Locladdr - Looplevl);
		jump(Endlab);
	}
	else if (Token == KLOOP) {
		Token = scan();
		semi();
		if (!Startlab) error("`LOOP' in wrong context", NULL);
		unstack(Locladdr - Looplevl);
		jump(Startlab);
	}
	else if (Token == KRETURN) {
		Token = scan();
		expr();
		semi();
		if (!Exitlab) error("`RETURN' in wrong context", NULL);
		pop();
		unstack(Locladdr-1);
		jump(Exitlab);
	}
	else if (Token == KHALT) {
		Token = scan();
		semi();
		genhalt();
	}
	else if (Token == SEMI) {
		Token = scan();
	}
	else {
		asg_or_call();
		semi();
	}
}


cleanup() {
	int	k, la;

	la = Locladdr;
	if ((k = findsym("*")) != -1) {
		St = k;
		Nt = Symbols[k].name - Names;
		Locladdr = Symbols[k].val;
	}
	if (la != Locladdr) unstack(la - Locladdr);
}


compound() {
	int	ot = Locladdr;

	newsym("*", 0, Locladdr, 0);
	Token = scan();
	while (Token == KVAR || Token == KCONST) {
		declaration(1);
	}
	stack(Locladdr - ot);
	while (Token != KEND) {
		if (feof(stdin)) fatal("unexpected EOF", NULL);
		stmt();
	}
	Token = scan();
	cleanup();
}


adjust(k, n) {
	n = n+3;
	while (k<St) {
		Symbols[k].val = -(n-Symbols[k].val);
		k++;
	}
}


fndecl() {
	int	p, k, lv, n = 0, addr = 2;

	p = findsym(Text);
	Exitlab = newlab();
	k = newsym(Text, 0, 0, FUNCT);
	lv = newsym("*", 0, 1, 0)+1;
	Token = scan();
	lparen();
	if (Token != RPAREN) for (;;) {
		if (Token == SYMBOL) {
			newsym(Text, 0, addr++, 0);
		}
		match(SYMBOL, "argument name expected");
		n++;
		if (Token != COMMA) break;
		Token = scan();
	}
	Symbols[k].size = n;
	if (p != -1 && (Symbols[p].flags & PROTO)) {
		if (Symbols[p].size != n)
			error("argument count mismatch", Symbols[k].name);
		Symbols[k].val = Symbols[p].val;
	}
	else {
		Symbols[k].val = newlab();
	}
	fnhdr(Symbols[k].name, Symbols[k].val);
	adjust(lv, n);
	rparen();
	stmt();
	cleanup();
	fnexit(Exitlab);
	Exitlab = 0;
}


fnproto() {
	int	k;

	Token = scan();
	for (;;) {
		if (Token != SYMBOL) {
			error("symbol expected", NULL);
			synch(SEMI); Token = scan();
			return;
		}
		k = newsym(Text, 0, 0, FUNCT|PROTO);
		Token = scan();
		lparen();
		Symbols[k].size = constval();
		Symbols[k].val = newlab();
		rparen();
		if (Token != COMMA) break;
		Token = scan();
	}
	semi();
}


declaration(local) {
	int	t, k;

	if (Token == KVAR || Token == KCONST) {
		t = Token;
		Token = scan();
		for (;;) {
			if (Token != SYMBOL) {
				error("symbol expected", NULL);
				synch(SEMI); Token = scan();
				return;
			}
			k = newsym(Text, 0, 0,
				(t==KCONST? CNST: 0) | (local? 0 : GLOBAL));
			Token = scan();
			if (t == KCONST) {
				equsign();
				Symbols[k].val = constval();
			}
			else {
				if (Token == LBRACK) {
					Token = scan();
					Symbols[k].size = constval();
					match(RBRACK, "missing ']'");
					if (Symbols[k].size < 1)
						error("bad array size",
							Symbols[k].name);
				}
				if (local) {
					Symbols[k].val = Locladdr;
					if (PUSHDOWN && Symbols[k].size)
						Symbols[k].val +=
							(Symbols[k].size-1);
					Locladdr += (Symbols[k].size?
						Symbols[k].size: 1);
				}
				else {
					Symbols[k].val = newlab();
					globdecl(Symbols[k].name,
						Symbols[k].size);
				}
			}
			if (Token != COMMA) break;
			Token = scan();
		}
		semi();
	}
	else if (Token == KDECL) {
		fnproto();
	}
	else if (Token == SYMBOL) {
		fndecl();
	}
	else {
		error("bad declaration", NULL);
		synch(0);
	}
}


external(s, n)
char	*s;
int	n;
{
	newsym(s, n, newlab(), FUNCT);
	extdecl(s);
}


init() {
	external("_open_", 2);
	external("_close_", 1);
	external("_erase_", 1);
	external("_select_", 2);
	external("_reads_", 2);
	external("_writes_", 1);
	external("_newline_", 0);
	external("_aton_", 1);
	external("_ntoa_", 2);
	external("_pack_", 2);
	external("_unpack_", 2);
	newsym("__version__", 0, 30, CNST|GLOBAL);
	Sym_equ = findop2("=");
	Sym_mod = findop2("mod");
}


main() {
	prologue();
	init();
	Token = scan();
	while (Token == KVAR || Token == KCONST || Token == SYMBOL ||
		Token == KDECL
	)
		declaration(0);
	if (Token != KDO) fatal("`DO' or declaration expected", NULL);
	start();
	compound();
	epilogue();
	return(Errcount? 1: 0);
}

